# -*- coding: utf-8 -*-
"""
Benchmarks for metrics.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyperf import Runner
from pyperf import perf_counter

from perfmetrics import metric
from perfmetrics import metricmethod
from perfmetrics import set_statsd_client
from perfmetrics import Metric
from perfmetrics.statsd import null_client

metricsampled_1 = Metric(rate=0.1)
metricsampled_9 = Metric(rate=0.999)

INNER_LOOPS = 1000

@metric
def func_with_metric():
    pass

@metricsampled_1
def func_with_metricsampled_1():
    pass

@metricsampled_9
def func_with_metricsampled_99():
    pass

def func_without_metric():
    pass

class AClass(object):

    @metricmethod
    def method_with_metric(self):
        pass

    def method_without_metric(self):
        pass


def _bench_call_func(loops, f):
    count = range(loops * INNER_LOOPS)
    t0 = perf_counter()
    for _ in count:
        f()
    t1 = perf_counter()
    return t1 - t0

##
# These four measure the overhead when there is no client installed
# First two are baselines.
##

def bench_a_call_func_without_metric(loops):
    return _bench_call_func(loops, func_without_metric)

def bench_a_call_method_without_metric(loops):
    return _bench_call_func(loops, AClass().method_without_metric)


def bench_call_func_with_metric(loops):
    return _bench_call_func(loops, func_with_metric)

def bench_call_method_with_metric(loops):
    return _bench_call_func(loops, AClass().method_with_metric)


##
# This measures the overhead of a trivial client
##

def _bench_call_func_with_client(loops,
                                 f=func_with_metric,
                                 client=null_client):
    set_statsd_client(client)
    result = _bench_call_func(loops, f)
    set_statsd_client(None)
    return result

def bench_call_func_with_null_client(loops):
    return _bench_call_func_with_client(loops)

##
# This measures the sampling overhead.
# It also uses a trivial client so we're not
# dependent on the order of tests.
##
def bench_call_sampled_1_func_with_null_client(loops):
    return _bench_call_func_with_client(loops, func_with_metricsampled_1, null_client)

def bench_call_sampled_99_func_with_null_client(loops):
    return _bench_call_func_with_client(loops, func_with_metricsampled_99, null_client)


##
# This measures actually sending the UDP packet
##
def bench_call_func_with_udp_client(loops):
    return _bench_call_func_with_client(
        loops,
        func_with_metric,
        'statsd://localhost:8125'
    )

def main():
    runner = Runner()
    for name, func in sorted([
            item for item in globals().items()
            if item[0].startswith('bench_')
    ]):
        runner.bench_time_func(name, func, inner_loops=INNER_LOOPS)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Helper functions.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] >= 3
PYPY = hasattr(sys, 'pypy_version_info')
WIN = sys.platform.startswith("win")
LINUX = sys.platform.startswith('linux')
OSX = sys.platform == 'darwin'


PURE_PYTHON = PYPY or os.getenv('PURE_PYTHON') or os.getenv("PERFMETRICS_PURE_PYTHON")

def import_c_accel(globs, cname):
    """
    Import the C-accelerator for the __name__
    and copy its globals.
    """

    name = globs.get('__name__')

    if not name or name == cname:
        # Do nothing if we're being exec'd as a file (no name)
        # or we're running from the C extension
        return # pragma: no cover


    if not PURE_PYTHON: # pragma: no cover
        import importlib
        import warnings
        with warnings.catch_warnings():
            # Python 3.7 likes to produce
            # "ImportWarning: can't resolve
            #   package from __spec__ or __package__, falling back on
            #   __name__ and __path__"
            # when we load cython compiled files. This is probably a bug in
            # Cython, but it doesn't seem to have any consequences, it's
            # just annoying to see and can mess up our unittests.
            warnings.simplefilter('ignore', ImportWarning)
            mod = importlib.import_module(cname)

        # By adopting the entire __dict__, we get a more accurate
        # __file__ and module repr, plus we don't leak any imported
        # things we no longer need.
        globs.clear()
        globs.update(mod.__dict__)

    if 'import_c_accel' in globs:
        del globs['import_c_accel']



# -*- coding: utf-8 -*-
"""
Optional WSGI integration.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .statsd import statsd_client_from_uri
from .clientstack import client_stack as statsd_client_stack
from .metric import Metric

def make_statsd_app(nextapp, _globals=None, statsd_uri=''):
    """
    Create a WSGI filter app that sets up Statsd for each request.

    If no *statsd_uri* is given, returns *nextapp* unchanged.

    .. versionchanged:: 3.0

       The returned app callable makes the statsd client that it
       uses available at the ``statsd_client`` attribute.
    """
    if not statsd_uri:
        # Disabled.
        return nextapp

    client = statsd_client_from_uri(statsd_uri)

    nextapp = Metric('perfmetrics.wsgi')(nextapp)
    def app(environ, start_response):
        statsd_client_stack.push(client)
        try:
            return nextapp(environ, start_response)
        finally:
            statsd_client_stack.pop()

    app.statsd_client = client

    return app



# -*- coding: utf-8 -*-
"""
Optional pyramid integration.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .statsd import statsd_client_from_uri
from .clientstack import client_stack as statsd_client_stack
from .metric import Metric

logger = __import__('logging').getLogger(__name__)

def includeme(config):
    """Pyramid configuration hook: activate the perfmetrics tween.

    A statsd_uri should be in the settings.
    """
    if config.registry.settings.get('statsd_uri'):
        config.add_tween('perfmetrics.tween')


def tween(handler, registry):
    """Pyramid tween that sets up a Statsd client for each request.
    """
    uri = registry.settings['statsd_uri']
    client = statsd_client_from_uri(uri)

    handler = Metric('perfmetrics.tween')(handler)
    def handle(request):
        statsd_client_stack.push(client)
        try:
            return handler(request)
        finally:
            statsd_client_stack.pop()

    return handle



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

class MockStatsdClient(object):
    def __init__(self):
        self.changes = []
        self.timings = []
        self.sentbufs = []

    def incr(self, stat, count=1, rate=1, buf=None,
             rate_applied=False):
        self.changes.append((stat, count, rate, buf, rate_applied))
        if buf is not None:
            buf.append('count_line')

    def timing(self, stat, ms, rate, buf=None,
               rate_applied=False):
        self.timings.append((stat, ms, rate, buf, rate_applied))
        if buf is not None:
            buf.append('timing_line')

    def sendbuf(self, buf):
        self.sentbufs.append(buf)


class TestMetric(unittest.TestCase):

    def setUp(self):
        self.statsd_client_stack.clear()

    tearDown = setUp

    @property
    def _class(self):
        from perfmetrics import Metric
        return Metric

    def _makeOne(self, *args, **kwargs):
        return self._class(*args, **kwargs)

    @property
    def statsd_client_stack(self):
        from perfmetrics import statsd_client_stack
        return statsd_client_stack

    def _add_client(self):
        client = MockStatsdClient()
        self.statsd_client_stack.push(client)
        return client

    def test_ctor_with_defaults(self):
        obj = self._makeOne()
        self.assertIsNone(obj.stat)
        self.assertEqual(obj.rate, 1)
        self.assertTrue(obj.count)
        self.assertTrue(obj.timing)
        from perfmetrics import _util
        if not _util.PURE_PYTHON: # pragma: no cover
            self.assertIn('_metric', str(obj))
        else:
            self.assertNotIn('_metric', str(obj))

    def test_ctor_with_options(self):
        obj = self._makeOne('spam.n.eggs', 0.1, count=False, timing=False)
        self.assertEqual(obj.stat, 'spam.n.eggs')
        self.assertEqual(obj.rate, 0.1)
        self.assertFalse(obj.count)
        self.assertFalse(obj.timing)

    def test_decorate_function(self):
        args = []
        metric = self._makeOne()

        @metric
        def spam(x, y=2):
            args.append((x, y))

        self.assertEqual(spam.__module__, __name__)
        self.assertEqual(spam.__name__, 'spam')

        # Call with no statsd client configured.
        spam(4, 5)
        self.assertEqual(args, [(4, 5)])
        del args[:]

        # Call with a statsd client configured.
        client = self._add_client()
        spam(6, 1)
        self.assertEqual(args, [(6, 1)])

        stat, delta, rate, buf, rate_applied = client.changes[0]
        self.assertEqual(stat, __name__ + '.spam')
        self.assertEqual(delta, 1)
        self.assertEqual(rate, 1)
        self.assertEqual(buf, ['count_line', 'timing_line'])
        self.assertTrue(rate_applied)

        self.assertEqual(len(client.timings), 1)
        stat, ms, rate, _buf, rate_applied = client.timings[0]
        self.assertEqual(stat, __name__ + '.spam.t')
        self.assertGreaterEqual(ms, 0)
        self.assertLess(ms, 10000)
        self.assertEqual(rate, 1)
        self.assertTrue(rate_applied)

        self.assertEqual(client.sentbufs, [['count_line', 'timing_line']])

    def test_decorate_method(self):
        args = []
        metricmethod = self._makeOne(method=True)

        class Spam(object):
            @metricmethod
            def f(self, x, y=2):
                args.append((self, x, y))

        self.assertEqual(Spam.f.__module__, __name__)
        self.assertEqual(Spam.f.__name__, 'f')

        # Call with no statsd client configured.
        spam = Spam()
        spam.f(4, 5)
        self.assertEqual(args, [(spam, 4, 5)])
        del args[:]

        # Call with a statsd client configured.
        client = self._add_client()
        spam.f(6, 1)
        self.assertEqual(args, [(spam, 6, 1)])

        self.assertEqual(len(client.changes), 1)
        __traceback_info__ = client.changes
        stat, delta, rate, buf, rate_applied = client.changes[0]
        self.assertEqual(stat, __name__ + '.Spam.f')
        self.assertEqual(delta, 1)
        self.assertEqual(rate, 1)
        self.assertEqual(buf, ['count_line', 'timing_line'])
        self.assertTrue(rate_applied)

        self.assertEqual(len(client.timings), 1)
        stat, ms, rate, _buf, rate_applied = client.timings[0]
        self.assertEqual(stat, __name__ + '.Spam.f.t')
        self.assertGreaterEqual(ms, 0)
        self.assertLess(ms, 10000)
        self.assertEqual(rate, 1)
        self.assertTrue(rate_applied)

        self.assertEqual(client.sentbufs, [['count_line', 'timing_line']])

    def test_decorate_can_change_timing(self):
        metric = self._makeOne()
        args = []
        @metric
        def spam(x, y=2):
            args.append((x, y))

        metric = spam.__wrapped__ if not hasattr(spam, 'metric_timing') else spam

        self.assertTrue(metric.metric_timing)
        self.assertTrue(metric.metric_count)
        self.assertEqual(1, metric.metric_rate)
        metric.metric_rate = 0
        metric.metric_timing = False
        metric.metric_count = False
        # Call with a statsd client configured.
        client = self._add_client()
        spam(6, 1)
        self.assertEqual(args, [(6, 1)])

        self.assertEqual(len(client.changes), 0)

    def test_decorate_method_can_change_timing(self):
        metricmethod = self._makeOne(method=True)
        args = []
        class Spam(object):
            @metricmethod
            def f(self, x, y=2):
                args.append((self, x, y))

        metric = Spam.f
        if not hasattr(metric, 'metric_timing'):
            metric = Spam.f.__wrapped__ # pylint:disable=no-member
        self.assertTrue(metric.metric_timing)
        self.assertTrue(metric.metric_count)
        self.assertEqual(1, metric.metric_rate)
        metric.metric_rate = 0
        metric.metric_timing = False
        metric.metric_count = False
        # Call with a statsd client configured.
        client = self._add_client()
        spam = Spam()
        spam.f(6, 1)
        self.assertEqual(args, [(spam, 6, 1)])

        self.assertEqual(len(client.changes), 0)


    def test_decorate_without_timing(self):
        args = []
        Metric = self._makeOne

        @Metric('spammy', rate=0.01, timing=False, random=lambda: 0.001)
        def spam(x, y=2):
            args.append((x, y))

        self.assertEqual(spam.__module__, __name__)
        self.assertEqual(spam.__name__, 'spam')

        # Call with no statsd client configured.
        spam(4, 5)
        self.assertEqual(args, [(4, 5)])
        del args[:]

        # Call with a statsd client configured.
        client = self._add_client()
        spam(6, 1)
        self.assertEqual(args, [(6, 1)])

        self.assertEqual(len(client.changes), 1)
        stat, delta, rate, buf, rate_applied = client.changes[0]
        self.assertEqual(stat, 'spammy')
        self.assertEqual(delta, 1)
        self.assertEqual(rate, 0.01)
        self.assertIsNone(buf)
        self.assertTrue(rate_applied)

        self.assertEqual(len(client.timings), 0)
        self.assertEqual(client.sentbufs, [])

    def test_decorate_without_count(self):
        args = []
        Metric = self._makeOne

        @Metric(count=False)
        def spam(x, y=2):
            args.append((x, y))

        self.assertEqual(spam.__module__, __name__)
        self.assertEqual(spam.__name__, 'spam')

        # Call with no statsd client configured.
        spam(4, 5)
        self.assertEqual(args, [(4, 5)])
        del args[:]

        # Call with a statsd client configured.
        client = self._add_client()
        spam(6, 1)
        self.assertEqual(args, [(6, 1)])
        self.assertEqual(client.changes, [])
        self.assertEqual(len(client.timings), 1)

        stat, ms, rate, buf, rate_applied = client.timings[0]
        self.assertEqual(stat, __name__ + '.spam.t')
        self.assertGreaterEqual(ms, 0)
        self.assertLess(ms, 10000)
        self.assertEqual(rate, 1)
        self.assertIsNone(buf)
        self.assertTrue(rate_applied)

        self.assertEqual(client.sentbufs, [])

    def test_decorate_with_neither_timing_nor_count(self):
        args = []
        Metric = self._makeOne

        @Metric(count=False, timing=False)
        def spam(x, y=2):
            args.append((x, y))

        # Call with no statsd client configured.
        spam(4, 5)
        self.assertEqual(args, [(4, 5)])
        del args[:]

        # Call with a statsd client configured.
        client = self._add_client()
        spam(6, 1)
        self.assertEqual(args, [(6, 1)])
        self.assertEqual(client.changes, [])
        self.assertEqual(len(client.timings), 0)

        self.assertEqual(client.sentbufs, [])

    def test_ignore_function_sample(self):
        args = []
        Metric = self._makeOne

        @Metric(rate=0.99, random=lambda: 0.999)
        def spam(x, y=2):
            args.append((x, y))
            return 77

        client = self._add_client()
        self.assertEqual(77, spam(6, 1))

        # The function was called
        self.assertEqual(args, [(6, 1)])

        # No packets were sent because the random value was too high.
        self.assertFalse(client.changes)
        self.assertFalse(client.timings)
        self.assertFalse(client.sentbufs)

    def test_as_context_manager_with_stat_name(self):
        args = []
        Metric = self._makeOne

        def spam(x, y=2):
            with Metric('thing-done'):
                args.append((x, y))

        # Call with no statsd client configured.
        spam(4, 5)
        self.assertEqual(args, [(4, 5)])
        del args[:]

        # Call with a statsd client configured.
        client = self._add_client()
        spam(6, 1)
        self.assertEqual(args, [(6, 1)])

        stat, delta, rate, buf, rate_applied = client.changes[0]
        self.assertEqual(stat, 'thing-done')
        self.assertEqual(delta, 1)
        self.assertEqual(rate, 1)
        self.assertEqual(buf, ['count_line', 'timing_line'])
        self.assertTrue(rate_applied)

        self.assertEqual(len(client.timings), 1)
        stat, ms, rate, _buf, rate_applied = client.timings[0]
        self.assertEqual(stat, 'thing-done.t')
        self.assertGreaterEqual(ms, 0)
        self.assertLess(ms, 10000)
        self.assertEqual(rate, 1)
        self.assertTrue(rate_applied)

        self.assertEqual(client.sentbufs, [['count_line', 'timing_line']])

    def test_ignore_context_manager_sample(self):
        args = []
        Metric = self._makeOne

        def spam(x, y=2):
            with Metric('thing-done', rate=0.99, random=lambda: 0.999):
                args.append((x, y))
                return 88

        client = self._add_client()
        self.assertEqual(88, spam(6, 716))

        # The function was called
        self.assertEqual(args, [(6, 716)])

        # No packets were sent because the random value was too high.
        self.assertFalse(client.changes)
        self.assertFalse(client.timings)
        self.assertFalse(client.sentbufs)

    def test_as_context_manager_without_stat_name(self):
        args = []
        Metric = self._makeOne

        def spam(x, y=2):
            with Metric():
                args.append((x, y))

        client = self._add_client()
        spam(6, 1)
        self.assertEqual(args, [(6, 1)])

        self.assertFalse(client.changes)
        self.assertFalse(client.timings)
        self.assertFalse(client.sentbufs)

    def test_as_context_manager_without_timing(self):
        args = []
        Metric = self._makeOne

        def spam(x, y=2):
            with Metric('thing-done', timing=False):
                args.append((x, y))

        client = self._add_client()
        spam(6, 1)
        self.assertEqual(args, [(6, 1)])

        self.assertEqual(len(client.changes), 1)

        stat, delta, rate, buf, rate_applied = client.changes[0]
        self.assertEqual(stat, 'thing-done')
        self.assertEqual(delta, 1)
        self.assertEqual(rate, 1)
        self.assertEqual(buf, ['count_line'])
        self.assertTrue(rate_applied)

        self.assertEqual(len(client.timings), 0)

        self.assertEqual(client.sentbufs, [['count_line']])

    def test_as_context_manager_without_count(self):
        args = []
        Metric = self._makeOne

        def spam(x, y=2):
            with Metric('thing-done', count=False):
                args.append((x, y))

        # Call with a statsd client configured.
        client = self._add_client()
        spam(6, 1)
        self.assertEqual(args, [(6, 1)])

        self.assertEqual(len(client.changes), 0)
        self.assertEqual(len(client.timings), 1)
        stat, ms, rate, _buf, rate_applied = client.timings[0]
        self.assertEqual(stat, 'thing-done.t')
        self.assertGreaterEqual(ms, 0)
        self.assertLess(ms, 10000)
        self.assertEqual(rate, 1)
        self.assertTrue(rate_applied)

        self.assertEqual(client.sentbufs, [['timing_line']])

    def test_as_context_manager_with_neither_count_nor_timing(self):
        args = []
        Metric = self._makeOne

        def spam(x, y=2):
            with Metric('thing-done', count=False, timing=False):
                args.append((x, y))

        # Call with a statsd client configured.
        client = self._add_client()
        spam(6, 1)
        self.assertEqual(args, [(6, 1)])

        self.assertEqual(len(client.changes), 0)
        self.assertEqual(len(client.timings), 0)
        self.assertEqual(client.sentbufs, [])

class SequenceDecorator(object):

    def __init__(self, *args, **kwargs):
        from perfmetrics import Metric
        from perfmetrics import MetricMod

        self.args = args
        self.kwargs = kwargs
        self.metric = Metric(*args, **kwargs)
        self.metricmod = MetricMod("%s")

    def __getattr__(self, name):
        return getattr(self.metric, name)

    def __call__(self, func):
        return self.metricmod(self.metric(func))

    def __enter__(self):
        self.metricmod.__enter__()
        self.metric.__enter__()

    def __exit__(self, t, v, tb):
        self.metric.__exit__(t, v, tb)
        self.metricmod.__exit__(t, v, tb)

    def __str__(self):
        return str(self.metricmod)


class TestMetricMod(TestMetric):

    def _makeOne(self, *args, **kwargs):
        return SequenceDecorator(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"


import functools

import unittest

from hamcrest import none
from hamcrest import assert_that
from hamcrest import contains_exactly as contains
from hamcrest import is_
from hamcrest import has_length
from hamcrest import calling
from hamcrest import raises

from ..matchers import is_counter
from ..matchers import is_timer
from ..matchers import is_gauge
from ..matchers import is_set

from ..observation import Observation as Metric


class TestMetricParsing(unittest.TestCase):

    def test_invalid_packet(self):
        packet = 'junk'
        assert_that(calling(functools.partial(Metric.make_all, packet)),
                    raises(ValueError))

        packet = 'foo|bar|baz|junk'
        assert_that(calling(functools.partial(Metric.make_all, packet)),
                    raises(ValueError))

        packet = 'gorets:1|c|0.1'
        assert_that(calling(functools.partial(Metric.make_all, packet)),
                    raises(ValueError))

    def test_counter(self):
        packet = 'gorets:1|c'
        metric = Metric.make_all(packet)

        assert_that(metric, has_length(1))
        metric = metric[0]

        assert_that(metric, is_counter('gorets', '1', none()))

    def test_sampled_counter(self):
        packet = 'gorets:1|c|@0.1'
        metric = Metric.make_all(packet)

        assert_that(metric, has_length(1))
        metric = metric[0]

        assert_that(metric, is_counter('gorets', '1', 0.1))

    def test_timer(self):
        packet = 'glork:320|ms'
        metric = Metric.make_all(packet)

        assert_that(metric, has_length(1))
        metric = metric[0]

        assert_that(metric, is_timer('glork', '320'))

    def test_set(self):
        packet = 'glork:3|s'
        metric = Metric.make_all(packet)

        assert_that(metric, has_length(1))
        metric = metric[0]

        assert_that(metric, is_set('glork', '3'))

    def test_gauge(self):
        packet = 'gaugor:+333|g'
        metric = Metric.make_all(packet)

        assert_that(metric, has_length(1))
        metric = metric[0]

        assert_that(metric, is_gauge('gaugor', '+333'))

    def test_multi_metric(self):
        packet = 'gorets:1|c\nglork:320|ms\ngaugor:333|g\nuniques:765|s'
        metrics = Metric.make_all(packet)
        assert_that(metrics, contains(is_counter(),
                                      is_timer(),
                                      is_gauge(),
                                      is_set()))

    def test_metric_string(self):
        metric = Metric.make('gaugor:+333|g')
        assert_that(str(metric), is_('gaugor:+333|g'))

        packet = 'gorets:1|c|@0.1'
        metric = Metric.make_all(packet)[0]
        metric = Metric.make_all(str(metric))[0]

        assert_that(metric, is_counter('gorets', '1', 0.1))

    def test_factory(self):
        metric = Metric.make('gaugor:+333|g')
        assert_that(str(metric), is_('gaugor:+333|g'))

        assert_that(calling(functools.partial(Metric.make, 'gaugor:+333|g\ngaugor:+333|g')),
                    raises(ValueError))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"


import unittest

from hamcrest import all_of
from hamcrest import assert_that
from hamcrest import contains_string
from hamcrest import is_not
from hamcrest import none

from hamcrest.core.string_description import StringDescription

from ..matchers import is_observation as is_metric
from ..matchers import is_counter
from ..matchers import is_gauge
from ..matchers import is_set
from ..matchers import is_timer

from ..observation import Observation as Metric


class TestIsMetric(unittest.TestCase):

    def setUp(self):
        self.counter = Metric.make('foo:1|c')
        self.timer = Metric.make('foo:100|ms|@0.1')
        self.set = Metric.make('foo:bar|s')
        self.gauge = Metric.make('foo:200|g')

    def test_is_metric(self):
        assert_that(self.counter, is_metric('c'))
        assert_that(self.counter, is_metric('c', 'foo'))
        assert_that(self.counter, is_metric('c', 'foo', '1'))
        assert_that(self.counter, is_metric('c', 'foo', '1', None))

    def test_non_metric(self):
        assert_that(object(), is_not(is_metric()))

    def test_is_counter(self):
        assert_that(self.counter, is_counter())

    def test_is_gauge(self):
        assert_that(self.gauge, is_gauge())

    def test_is_set(self):
        assert_that(self.set, is_set())

    def test_is_timer(self):
        assert_that(self.timer, is_timer())

    def test_bad_kind(self):
        assert_that(self.counter, is_not(is_timer()))

    def test_bad_name(self):
        assert_that(self.counter, is_not(is_counter('bar')))

    def test_bad_value(self):
        assert_that(self.counter, is_not(is_counter('foo', '2')))

    def test_bad_sampling(self):
        assert_that(self.timer, is_not(is_counter('foo', '100', None)))

    def test_failure_error(self):
        desc = StringDescription()
        matcher = is_counter('foo', '1', 0.1)
        matcher.describe_to(desc)
        desc = str(desc)
        # Strip internal newlines, which vary between
        # hamcrest versions. Also, the exact text varies too;
        # beginning with 1.10, the has_properties matcher we use internally is
        # less verbose by default and so we get a string like:
        #   an object with properties kind matching c and name matching foo
        # where before it was something like
        #  an object with property kind matching c
        #  and an object with property name matching foo

        desc = desc.replace('\n', '')
        assert_that(
            desc,
            all_of(
                contains_string(
                    "(an instance of Observation and "),
                contains_string(
                    "'kind' matching 'c'"),
                contains_string(
                    "'name' matching 'foo'"),
                contains_string(
                    "'value' matching '1'"),
                contains_string(
                    "'sampling_rate' matching <0.1>"
                )
            )
        )


    def test_components_can_be_matchers(self):
        assert_that(self.counter, is_metric('c', 'foo', '1', none()))
        assert_that(self.timer, is_not(is_metric('ms', 'foo', '100', none())))



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from hamcrest.core.base_matcher import BaseMatcher


# XXX These matchers are temporary for Python 3.10 pending zodbpickle 2.2
# (nti.testing can't be imported on 3.10 before that version)

try:
    from nti.testing.matchers import is_true # pylint:disable=unused-import
except ImportError: # pragma: no cover
    class BoolMatcher(BaseMatcher):
        def __init__(self, value):
            super(BoolMatcher, self).__init__()
            self.value = value

        def _matches(self, item):
            return bool(item) == self.value

        def describe_to(self, description):
            description.append_text('object with bool() value ').append(str(self.value))

        def __repr__(self):
            return 'object with bool() value ' + str(self.value)

    def is_true():
        """
        Matches an object with a true boolean value.
        """
        return BoolMatcher(True)

try:
    from nti.testing.matchers import implements # pylint:disable=unused-import
except ImportError: # pragma: no cover
    class Implements(BaseMatcher):

        def __init__(self, iface):
            super(Implements, self).__init__()
            self.iface = iface

        def _matches(self, item):
            return self.iface.implementedBy(item)

        def describe_to(self, description):
            description.append_text('object implementing')
            description.append_description_of(self.iface)

    def implements(iface):
        """
        Matches if the object implements (is a factory for) the given
        interface.

        .. seealso:: :meth:`zope.interface.interfaces.ISpecification.implementedBy`
        """
        return Implements(iface)

try:
    from nti.testing.matchers import validly_provides # pylint:disable=unused-import
except ImportError: # pragma: no cover
    from zope.schema import ValidationError
    from zope.schema import getValidationErrors
    from zope.interface.exceptions import Invalid
    from zope.interface.verify import verifyObject
    import hamcrest

    class VerifyProvides(BaseMatcher):

        def __init__(self, iface):
            super(VerifyProvides, self).__init__()
            self.iface = iface

        def _matches(self, item):
            try:
                verifyObject(self.iface, item)
            except Invalid:
                return False
            else:
                return True

        def describe_to(self, description):
            description.append_text('object verifiably providing ').append_description_of(
                self.iface)

        def describe_mismatch(self, item, mismatch_description):
            md = mismatch_description

            try:
                verifyObject(self.iface, item)
            except Invalid as x:
                # Beginning in zope.interface 5, the Invalid exception subclasses
                # like BrokenImplementation, DoesNotImplement, etc, all typically
                # have a much nicer error message than they used to, better than we
                # were producing. This is especially true now that MultipleInvalid
                # is a thing.
                x = str(x).strip()

                md.append_text("Using class ").append_description_of(type(item)).append_text(' ')
                if x.startswith('The object '):
                    x = x[len("The object "):]
                    x = 'the object ' + x
                x = x.replace('\n    ', '\n          ')
                md.append_text(x)


    def verifiably_provides(*ifaces):
        """
        Matches if the object verifiably provides the correct interface(s),
        as defined by :func:`zope.interface.verify.verifyObject`. This means having
        the right attributes
        and methods with the right signatures.

        .. note:: This does **not** test schema compliance. For that
            (stricter) test, see :func:`validly_provides`.
        """
        if len(ifaces) == 1:
            return VerifyProvides(ifaces[0])

        return hamcrest.all_of(*[VerifyProvides(x) for x in ifaces])

    class VerifyValidSchema(BaseMatcher):
        def __init__(self, iface):
            super(VerifyValidSchema, self).__init__()
            self.iface = iface

        def _matches(self, item):
            errors = getValidationErrors(self.iface, item)
            return not errors

        def describe_to(self, description):
            description.append_text('object validly providing ').append(str(self.iface))

        def describe_mismatch(self, item, mismatch_description):
            x = None
            md = mismatch_description
            md.append_text(str(type(item)))

            errors = getValidationErrors(self.iface, item)

            for attr, exc in errors:
                try:
                    raise exc
                except ValidationError:
                    md.append_text(' has attribute "')
                    md.append_text(attr)
                    md.append_text('" with error "')
                    md.append_text(repr(exc))
                    md.append_text('"\n\t ')
                except Invalid as x: # pragma: no cover
                    md.append_text(str(x))

    def validly_provides(*ifaces):
        """
        Matches if the object verifiably and validly provides the given
        schema (interface(s)).

        Verification is done with :mod:`zope.interface` and
        :func:`verifiably_provides`, while validation is done with
        :func:`zope.schema.getValidationErrors`.
        """
        if len(ifaces) == 1:
            the_schema = ifaces[0]
            return hamcrest.all_of(verifiably_provides(the_schema), VerifyValidSchema(the_schema))

        prov = verifiably_provides(*ifaces)
        valid = [VerifyValidSchema(x) for x in ifaces]

        return hamcrest.all_of(prov, *valid)



# -*- coding: utf-8 -*-
"""
Statsd client implementations.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import random
import socket

try:
    # Python 3
    from urllib.parse import urlsplit
    from urllib.parse import parse_qsl
    from urllib.parse import uses_query

    basestring = str

except ImportError: # pragma: no cover
    # Python 2
    from urlparse import urlsplit
    from urlparse import parse_qsl
    from urlparse import uses_query

from .interfaces import IStatsdClient
from .interfaces import implementer

logger = logging.getLogger(__name__)

__all__ = [
    'StatsdClient',
    'StatsdClientMod',
    'NullStatsdClient',
    'statsd_client_from_uri',
]

if 'statsd' not in uses_query:  # pragma: no cover
    uses_query.append('statsd')

def statsd_client_from_uri(uri):
    """
    Create and return :class:`perfmetrics.statsd.StatsdClient`.

    A typical URI is ``statsd://localhost:8125``. An optional query
    parameter is ``prefix``. The default prefix is an empty string.

    """
    parts = urlsplit(uri)
    if parts.scheme != 'statsd':
        raise ValueError("URI scheme not supported: %s" % uri)

    kw = {}
    if parts.query:
        kw.update(parse_qsl(parts.query))
    return StatsdClient(parts.hostname, parts.port, **kw)


@implementer(IStatsdClient)
class StatsdClient(object):
    """
    Send packets to statsd.

    Default implementation of :class:`perfmetrics.interfaces.IStatsdClient`.

    Derived from statsd.py by Steve Ivy <steveivy@gmail.com>.
    """

    def __init__(self, host='localhost', port=8125, prefix=''):
        # Resolve the host name early.
        info = socket.getaddrinfo(host, int(port), 0, socket.SOCK_DGRAM)
        family, socktype, proto, _canonname, addr = info[0]
        self.addr = addr
        self.log = logger
        self.udp_sock = socket.socket(family, socktype, proto)
        self.random = random.random  # Testing hook
        if prefix and not prefix.endswith('.'):
            prefix = prefix + '.'
        self.prefix = prefix

    def close(self):
        """
        See :meth:`perfmetrics.interfaces.IStatsdClient.close`.

        .. versionadded:: 3.0
        """
        if self.udp_sock:
            self.udp_sock.close()
            self.udp_sock = None

    def timing(self, stat, value, rate=1, buf=None, rate_applied=False):
        """
        See :meth:`perfmetrics.interfaces.IStatsdClient.timing`.

        """
        if rate >= 1 or rate_applied or self.random() < rate:
            s = '%s%s:%d|ms' % (self.prefix, stat, value)
            if buf is None:
                self._send(s)
            else:
                buf.append(s)

    def gauge(self, stat, value, rate=1, buf=None, rate_applied=False):
        """
        See :meth:`perfmetrics.interfaces.IStatsdClient.gauge`.
        """
        if rate >= 1 or rate_applied or self.random() < rate:
            s = '%s%s:%s|g' % (self.prefix, stat, value)
            if buf is None:
                self._send(s)
            else:
                buf.append(s)

    def incr(self, stat, count=1, rate=1, buf=None, rate_applied=False):
        """
        See :meth:`perfmetrics.interfaces.IStatsdClient.incr`.
        """
        if rate >= 1:
            s = '%s%s:%s|c' % (self.prefix, stat, count)
        elif rate_applied or self.random() < rate:
            s = '%s%s:%s|c|@%s' % (self.prefix, stat, count, rate)
        else:
            return

        if buf is None:
            self._send(s)
        else:
            buf.append(s)

    def decr(self, stat, count=1, rate=1, buf=None, rate_applied=False):
        """
        See :meth:`perfmetrics.interfaces.IStatsdClient.decr`.
        """
        self.incr(stat, -count, rate=rate, buf=buf, rate_applied=rate_applied)

    def set_add(self, stat, value, rate=1, buf=None, rate_applied=False):
        """
        See :meth:`perfmetrics.interfaces.IStatsdClient.set_add`.
        """
        if rate >= 1 or rate_applied or self.random() < rate:
            s = '%s%s:%s|s' % (self.prefix, stat, value)
            if buf is None:
                self._send(s)
            else:
                buf.append(s)

    def _send(self, data):
        """Send a UDP packet containing a string."""
        try:
            self.udp_sock.sendto(data.encode('ascii'), self.addr)
        except IOError:
            self.log.exception("Failed to send UDP packet")

    def sendbuf(self, buf):
        """
        See :meth:`perfmetrics.interfaces.IStatsdClient.sendbuf`.
        """
        if buf:
            self._send('\n'.join(buf))


@implementer(IStatsdClient)
class StatsdClientMod(object):
    """
    Wrap `StatsdClient`, modifying all stat names in context.

    .. versionchanged:: 3.0

       The wrapped object's attributes are now accessible on this object.

       This object now uses ``__slots__``.
    """

    __slots__ = (
        '_wrapped',
        'format',
    )

    def __init__(self, wrapped, format):
        self._wrapped = wrapped
        self.format = format

    def close(self):
        self._wrapped.close()

    def __getattr__(self, name):
        return getattr(self._wrapped, name)

    def __setattr__(self, name, value):
        if name in self.__slots__:
            object.__setattr__(self, name, value)
        else:
            setattr(self._wrapped, name, value)

    def timing(self, stat, *args, **kw):
        self._wrapped.timing(self.format % stat, *args, **kw)

    def gauge(self, stat, *args, **kw):
        self._wrapped.gauge(self.format % stat, *args, **kw)

    def incr(self, stat, *args, **kw):
        self._wrapped.incr(self.format % stat, *args, **kw)

    def decr(self, stat, *args, **kw):
        self._wrapped.decr(self.format % stat, *args, **kw)

    def set_add(self, stat, *args, **kw):
        self._wrapped.set_add(self.format % stat, *args, **kw)

    def sendbuf(self, buf):
        self._wrapped.sendbuf(buf)


@implementer(IStatsdClient)
class NullStatsdClient(object):
    """No-op statsd client."""

    def close(self):
        """Does nothing."""

    def timing(self, stat, *args, **kw):
        """Does nothing."""

    def gauge(self, stat, *args, **kw):
        """Does nothing."""

    def incr(self, stat, *args, **kw):
        """Does nothing."""

    def decr(self, stat, *args, **kw):
        """Does nothing."""

    def set_add(self, stat, value, *args, **kw):
        """Does nothing."""

    def sendbuf(self, buf):
        """Does nothing"""


null_client = NullStatsdClient()



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest


class MockRegistry(object):
    def __init__(self, statsd_uri=None):
        self.settings = {}
        if statsd_uri:
            self.settings['statsd_uri'] = statsd_uri

class MockConfig(object):
    def __init__(self, statsd_uri=None):
        self.registry = MockRegistry(statsd_uri)
        self.tweens = []

    def add_tween(self, name):
        self.tweens.append(name)


class Test_includeme(unittest.TestCase):

    def _call(self, config):
        from perfmetrics import includeme
        includeme(config)

    def _make_config(self, statsd_uri):
        return MockConfig(statsd_uri)

    def test_without_statsd_uri(self):
        config = self._make_config(None)
        self._call(config)
        self.assertFalse(config.tweens)

    def test_with_statsd_uri(self):
        config = self._make_config('statsd://localhost:9999')
        self._call(config)
        self.assertEqual(config.tweens, ['perfmetrics.tween'])


class Test_tween(unittest.TestCase):

    def setUp(self):
        from perfmetrics import set_statsd_client, statsd_client_stack
        set_statsd_client(None)
        statsd_client_stack.clear()

    def _call(self, handler, registry):
        from perfmetrics import tween
        return tween(handler, registry)

    def _make_registry(self, statsd_uri):
        return MockRegistry(statsd_uri)

    def test_call_tween(self):
        clients = []

        def dummy_handler(_request):
            from perfmetrics import statsd_client
            client = statsd_client()
            self.addCleanup(client.close)
            clients.append(client)
            return 'ok!'

        registry = self._make_registry('statsd://localhost:9999')
        tween = self._call(dummy_handler, registry)
        response = tween(object())
        self.assertEqual(response, 'ok!')
        self.assertEqual(len(clients), 1)
        from perfmetrics.statsd import StatsdClient
        self.assertIsInstance(clients[0], StatsdClient)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

from hamcrest.core.matcher import Matcher
from hamcrest.core.base_matcher import BaseMatcher

from hamcrest import all_of
from hamcrest import instance_of
from hamcrest import is_
from hamcrest import has_properties
from hamcrest import none


from .observation import OBSERVATION_KIND_COUNTER as METRIC_COUNTER_KIND
from .observation import OBSERVATION_KIND_GAUGE as METRIC_GAUGE_KIND
from .observation import OBSERVATION_KIND_SET as METRIC_SET_KIND
from .observation import OBSERVATION_KIND_TIMER as METRIC_TIMER_KIND

from .observation import Observation

__all__ = [
    'is_observation',
    'is_counter',
    'is_gauge',
    'is_set',
    'is_timer',
]

_marker = object()

_metric_kind_display_name = {
    'c': 'counter',
    'g': 'gauge',
    'ms': 'timer',
    's': 'set'
}


class IsMetric(BaseMatcher):

    # See _matches()
    _force_one_description = False

    def __init__(self, kwargs):
        matchers = {}
        for key, value in kwargs.items():
            if value is None:
                value = none()
            elif key == 'sampling_rate':
                # This one is special, it doesn't get
                # to be a string, it's kept as a number.
                value = is_(value)
            elif not isinstance(value, Matcher):
                value = str(value)
            matchers[key] = value

        # Beginning in 1.10 and up through at least 2.0.2,
        # has_properties con no longer be called with an empty dictionary
        # without creating a KeyError from popitem().
        self._matcher = all_of(
            instance_of(Observation),
            has_properties(**matchers) if matchers else instance_of(Observation),
        )
        if self._force_one_description:
            self.__patch_matcher()

    def __patch_matcher(self):
        # This is tightly coupled to the structure of the matcher we create.
        self._matcher.describe_all_mismatches = False
        has_prop_matcher = self._matcher.matchers[1]
        if hasattr(has_prop_matcher, 'matcher'):
            has_prop_matcher.matcher.describe_all_mismatches = False

    def _matches(self, item):
        # On PyHamcrest == 1.10.x (last release for Python 2)
        # There's a bug such that you can't call AllOf.matches without
        # passing in a description without getting
        # ``AttributeError: None has no append_text``
        # So we pass one, even though it gets thrown away.
        # XXX: Remove this workaround when we only support PyHamcrest 2.
        # See https://github.com/hamcrest/PyHamcrest/issues/130
        try:
            return self._matcher.matches(item)
        except AttributeError as ex:
            if 'append_text' not in str(ex): # pragma: no cover
                raise
            IsMetric._force_one_description = True
            self.__patch_matcher()
            return self._matcher.matches(item)

    def describe_to(self, description):
        self._matcher.describe_to(description)

    def describe_mismatch(self, item, mismatch_description):
        mismatch_description.append_text('was ').append_text(repr(item))

def _is_metric(args, kwargs):
    for arg_ix, name in enumerate((
            'kind',
            'name',
            'value',
            'sampling_rate'
    )):
        if name in kwargs or len(args) <= arg_ix:
            continue
        kwargs[name] = args[arg_ix]

    return IsMetric(kwargs)


def is_observation(*args, **kwargs):
    """
    is_observation(*, kind, name, value, sampling_rate) -> matcher

    A hamcrest matcher that validates the specific parts of a `~.Observation`.
    All arguments are optional and can be provided by name or position.

    :keyword str kind: A hamcrest matcher or string that matches the kind for this metric
    :keyword str name: A hamcrest matcher or string that matches the name for this metric
    :keyword str value: A hamcrest matcher or string that matches the value for this metric
    :keyword float sampling_rate: A hamcrest matcher or
        number that matches the sampling rate this metric was collected with
    """
    return _is_metric(args, kwargs)


def is_counter(*args, **kwargs):
    """
    is_counter(*, name, value, sampling_rate) -> matcher

    A hamcrest matcher validating the parts of a counter `~.Observation`.

    .. seealso:: `is_metric`
    """
    kwargs['kind'] = METRIC_COUNTER_KIND
    args = (None,) + args
    return _is_metric(args, kwargs)


def is_gauge(*args, **kwargs):
    """
    is_gauge(*, name, value, sampling_rate) -> matcher

    A hamcrest matcher validating the parts of a gauge `~.Observation`

    .. seealso:: `is_metric`
    """
    kwargs['kind'] = METRIC_GAUGE_KIND
    args = (None,) + args
    return _is_metric(args, kwargs)


def is_timer(*args, **kwargs):
    """
    is_timer(*, name, value, sampling_rate) -> matcher

    A hamcrest matcher validating the parts of a timer `~.Observation`

    .. seealso:: `is_metric`
    """
    kwargs['kind'] = METRIC_TIMER_KIND
    args = (None,) + args
    return _is_metric(args, kwargs)

def is_set(*args, **kwargs):
    """
    is_set(*, name, value, sampling_rate) -> matcher

    A hamcrest matcher validating the parts of a set `~.Observation`

    .. seealso:: `is_metric`
    """
    kwargs['kind'] = METRIC_SET_KIND
    args = (None,) + args
    return _is_metric(args, kwargs)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest


class Test_statsd_config_functions(unittest.TestCase):

    def setUp(self):
        from perfmetrics import set_statsd_client
        set_statsd_client(None)

    tearDown = setUp

    def test_unconfigured(self):
        from perfmetrics import statsd_client
        self.assertIsNone(statsd_client())

    def test_configured_with_uri(self):
        from perfmetrics import set_statsd_client
        set_statsd_client('statsd://localhost:8125')
        from perfmetrics import StatsdClient
        from perfmetrics import statsd_client
        client = statsd_client()
        self.addCleanup(client.close)
        self.assertIsInstance(client, StatsdClient)

    def test_configured_with_other_client(self):
        other_client = object()
        from perfmetrics import set_statsd_client
        set_statsd_client(other_client)
        from perfmetrics import statsd_client
        self.assertIs(statsd_client(), other_client)


class Test_statsd_client_from_uri(unittest.TestCase):

    def _call(self, uri):
        from perfmetrics import statsd_client_from_uri
        client = statsd_client_from_uri(uri)
        self.addCleanup(client.close)
        return client

    def test_local_uri(self):
        client = self._call('statsd://localhost:8129')
        self.assertIsNotNone(client.udp_sock)

    def test_unsupported_uri(self):
        with self.assertRaises(ValueError):
            self._call('http://localhost:8125')

    def test_with_custom_prefix(self):
        client = self._call('statsd://localhost:8129?prefix=spamalot')
        self.assertEqual(client.prefix, 'spamalot.')


class TestDisabledDecorators(unittest.TestCase):

    def setUp(self):
        import sys
        import os
        __import__('perfmetrics')
        self.old_mod = sys.modules['perfmetrics']

        del sys.modules['perfmetrics']
        os.environ['PERFMETRICS_DISABLE_DECORATOR'] = '1'
        self.perfmetrics = __import__("perfmetrics")
        del os.environ['PERFMETRICS_DISABLE_DECORATOR']

    def tearDown(self):
        import sys
        sys.modules['perfmetrics'] = self.old_mod

    def test_metric(self):

        def func():
            'does nothing'

        func2 = self.perfmetrics.metric(func)
        self.assertIs(func, func2)

    def test_metricmethod(self):

        def func():
            'does nothing'

        func2 = self.perfmetrics.metricmethod(func)
        self.assertIs(func, func2)

    def test_Metric(self):

        def func():
            'does nothing'

        func2 = self.perfmetrics.Metric()(func)
        self.assertIs(func, func2)

    def test_MetricMod(self):

        def func():
            'does nothing'

        func2 = self.perfmetrics.MetricMod('%s')(func)
        self.assertIs(func, func2)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading

from .statsd import statsd_client_from_uri

string_types = (str,)
if str is bytes: # pragma: no cover
    string_types += (unicode,) # pylint:disable=undefined-variable

class ClientStack(threading.local):
    """
    Thread local stack of StatsdClients.

    Applications and tests can either set the global statsd client using
    perfmetrics.set_statsd_client() or set a statsd client for each thread
    using statsd_client_stack.push()/.pop()/.clear().

    This is like pyramid.threadlocal but it handles the default differently.
    """

    default = None

    def __init__(self):
        threading.local.__init__(self) # pylint:disable=non-parent-init-called
        self.stack = []

    def get(self):
        """
        Return the current StatsdClient for the thread.

        Returns the thread-local client if there is one, or the global
        client if there is one, or None.
        """
        stack = self.stack
        return stack[-1] if stack else self.default

    def push(self, obj):
        self.stack.append(obj)

    def pop(self):
        stack = self.stack
        if stack:
            return stack.pop()

    def clear(self):
        del self.stack[:]


client_stack = ClientStack()

# Just expose the bound method, don't wrap it,
# for speed.
statsd_client = client_stack.get


def set_statsd_client(client_or_uri):
    """
    Set the global StatsdClient.

    The ``client_or_uri`` can be a StatsdClient, a ``statsd://`` URI,
    or None.

    Note that when the perfmetrics module is imported, it
    looks for the ``STATSD_URI`` environment variable and calls
    `set_statsd_client` if that variable is found.
    """
    if isinstance(client_or_uri, string_types):
        client = statsd_client_from_uri(client_or_uri)
    else:
        client = client_or_uri
    ClientStack.default = client



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904

import doctest
import os
import unittest

from hamcrest import assert_that
from hamcrest import contains_exactly as contains

from hamcrest import has_length
from hamcrest import has_properties
from hamcrest import has_property

from ...tests import is_true

from ...tests.test_statsd import TestBasics

from .. import FakeStatsDClient as MockStatsDClient

from ..observation import OBSERVATION_KIND_COUNTER as METRIC_COUNTER_KIND
from ..observation import OBSERVATION_KIND_GAUGE as METRIC_GAUGE_KIND
from ..observation import OBSERVATION_KIND_SET as METRIC_SET_KIND
from ..observation import OBSERVATION_KIND_TIMER as METRIC_TIMER_KIND

class TestMockStatsDClient(TestBasics):

    _class = MockStatsDClient

    def setUp(self):
        self.client = self._makeOne()

    def test_true_initially(self):
        assert_that(self.client, is_true())

    def test_tracks_metrics(self):
        self.client.incr('mycounter')
        self.client.gauge('mygauge', 5)
        self.client.timing('mytimer', 3003)
        self.client.set_add('myset', 42)

        assert_that(self.client, has_length(4))

        counter, gauge, timer, Set = self.client.observations

        assert_that(counter, has_properties('name', 'mycounter',
                                            'value', '1',
                                            'kind', METRIC_COUNTER_KIND))

        assert_that(gauge, has_properties('name', 'mygauge',
                                          'value', '5',
                                          'kind', METRIC_GAUGE_KIND))

        assert_that(timer, has_properties(
            'name', 'mytimer',
            'value', '3003',
            'kind', METRIC_TIMER_KIND
        ))

        assert_that(Set, has_properties(
            'name', 'myset',
            'value', '42',
            'kind', METRIC_SET_KIND
        ))

    def test_clear(self):
        self.client.incr('mycounter')
        assert_that(self.client, has_length(1))
        assert_that(self.client, is_true())

        self.client.clear()
        assert_that(self.client, has_length(0))
        assert_that(self.client, is_true())

    def test_tracks_multimetrics(self):
        packet = 'gorets:1|c\nglork:320|ms\ngaugor:333|g\nuniques:765|s'
        self.client._send(packet)

        assert_that(self.client, has_length(4))
        assert_that(self.client.packets, contains(packet))

        assert_that(self.client.observations,
                    contains(has_property('kind', METRIC_COUNTER_KIND),
                             has_property('kind', METRIC_TIMER_KIND),
                             has_property('kind', METRIC_GAUGE_KIND),
                             has_property('kind', METRIC_SET_KIND)))

def test_suite():
    root = this_dir = os.path.dirname(os.path.abspath(__file__))
    while not os.path.exists(os.path.join(root, 'setup.py')):
        prev, root = root, os.path.dirname(root)
        if root == prev:
            # We seem to be installed out of tree. Are we working in the
            # right directory at least?
            if os.path.exists('setup.py'): # pragma: no cover
                root = os.path.dirname(os.path.abspath('setup.py'))
                break
            # Let's avoid infinite loops at root
            raise AssertionError('could not find my setup.py')
    docs = os.path.join(root, 'docs')
    testing_rst = os.path.join(docs, 'testing.rst')

    optionflags = (
        doctest.NORMALIZE_WHITESPACE
        | doctest.ELLIPSIS
        | doctest.IGNORE_EXCEPTION_DETAIL
    )

    # Can't pass absolute paths to DocFileSuite, needs to be
    # module relative
    testing_rst = os.path.relpath(testing_rst, this_dir)

    return unittest.TestSuite((
        unittest.defaultTestLoader.loadTestsFromName(__name__),
        doctest.DocFileSuite(
            testing_rst,
            optionflags=optionflags
        ),
    ))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

#: The statsd metric kind for Gauges
OBSERVATION_KIND_GAUGE = 'g'

#: The statsd metric kind for Counters
OBSERVATION_KIND_COUNTER = 'c'

#: The statsd metric kind for Sets
OBSERVATION_KIND_SET = 's'

#: The statsd metric kind for Timers
OBSERVATION_KIND_TIMER = 'ms'


def _parse_sampling_data(data):
    """
    Parses sampling rate from the provided packet part *data*.

    Raises a `ValueError` if the packet part is invalid sampling data
    """
    if not data.startswith('@'):
        raise ValueError('Expected "@" in sampling data. %s' % data)
    return float(data[1:])


def _as_metric(metric_data):
    """
    Parses a single metric packet, *metric_data*, in to a `Metric`.

    Metrics take the form of <name>:<value>|<type>(|@<sampling_rate>)

    A `ValueError` is raised for invalid data
    """

    sampling = None
    name = None
    value = None
    kind = None
    parts = metric_data.split('|')
    if len(parts) < 2 or len(parts) > 3:
        raise ValueError('Unexpected metric data %s. Wrong number of parts' % metric_data)

    if len(parts) == 3:
        sampling_data = parts.pop(-1)
        sampling = _parse_sampling_data(sampling_data)

    kind = parts[1]
    name, value = parts[0].split(':')

    return Observation(name, value, kind, sampling_rate=sampling)


def _as_metrics(data):
    """
    Parses the statsd *data* packet to a _list_ of metrics.

    .. seealso:: https://github.com/etsy/statsd/blob/master/docs/metric_types.md
    """
    metrics = []

    # Multi metric packets are seperated by newlines
    for metric_data in data.split('\n'):
        metrics.append(_as_metric(metric_data))
    return metrics


class Observation(object):
    """
    The representation of a single statsd metric.
    """

    #: The metric name
    name = None

    #: The value provided for the metric
    value = None

    #: The statsd code for the type of metric. e.g. one of the ``METRIC_*_KIND`` constants
    kind = None

    #: The rate with which this event has been sampled from (optional)
    sampling_rate = None

    def __init__(self, name, value, kind, sampling_rate=None):
        self.name = name
        self.value = value
        self.sampling_rate = sampling_rate
        self.kind = kind

    @classmethod
    def make(cls, packet):
        """
        Creates a metric from the provided statsd *packet*.

        :raises ValueError: if *packet* is a multi metric packet or
                 otherwise invalid.
        """
        metrics = cls.make_all(packet)
        if len(metrics) != 1:
            raise ValueError('Must supply a single metric packet. %s supplied' % packet)
        return metrics[0]

    @classmethod
    def make_all(cls, packet):
        """
        Makes a list of metrics from the provided statsd *packet*.

        Like `make` but supports multi metric packets
        """
        return _as_metrics(packet)

    def __str__(self):
        sampling_string = '|@%g' % self.sampling_rate if self.sampling_rate is not None else ''
        return '%s:%s|%s%s' % (self.name, self.value, self.kind, sampling_string)

    def __repr__(self):
        return "%s(name=%r, value=%r, kind=%r, sampling_rate=%r)" % (
            self.__class__.__name__,
            self.name, self.value, self.kind, self.sampling_rate
        )



# definitions for metric.py

import cython

cdef time
cdef MethodType
cdef WeakKeyDictionary
cdef functools
cdef stdrandom

cdef statsd_client
cdef statsd_client_stack
cdef StatsdClientMod
cdef null_client

cdef class _MethodLikeMixin(object):
    pass

cdef class _AbstractMetricImpl(_MethodLikeMixin):
    cdef public bint metric_timing
    cdef public bint metric_count
    cdef public double metric_rate
    cdef  f
    cdef str timing_format
    cdef public __wrapped__
    cdef dict __dict__

    cdef str _compute_stat(self, tuple args)


cdef class _GivenStatMetricImpl(_AbstractMetricImpl):
    cdef readonly str stat_name

cdef class _MethodMetricImpl(_AbstractMetricImpl):

    cdef klass_dict


cdef class Metric(object):
    cdef public double rate
    cdef double start
    cdef public bint method
    cdef public bint count
    cdef public bint timing
    cdef public str stat
    cdef public str timing_format
    cdef random
    cdef dict __dict__
